// Stage3D Shoot-em-up Tutorial Part 5
// by Christer Kaitila - www.mcfunkypants.com

// Entity.as
// The Entity class will eventually hold all game-specific entity stats
// for the spaceships, bullets and effects in our game.
// It stores a reference to a gpu sprite and a few demo properties.
// This is where you would add hit points, weapons, ability scores, etc.
// This class handles any AI (artificial intelligence) for enemies as well.

package
{
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	public class Entity
	{
		// v5 stats used by the player entity
		// but could optionally be used by some enemies too
		public var health:int = 100;
		// how many reserve ships you have until game over
		public var lives:int = 3;
		// matches the score on the GUI
		public var score:int = 0;
		// what level the player is on
		public var level:int = 0;
		// when you get hit, for a short while you are impervious
		// to damage (otherwise nearby ships could add up to instant death)
		public var invulnerabilityTimeLeft:Number = 0;
		public var invulnerabilitySecsWhenHit:Number = 4;
		// when you die, change levels or get to a game over state,
		// this is the period of time that the GUI tells you
		// before switching to the new state
		public var transitionTimeLeft:Number = 0;
		public var transitionSeconds:Number = 5;
		// how much damage the player will take when getting hit
		// since the player has 100HP, 49 means you can be hit 3x
		public var damage:int = 49;
		
		// used for sprite animation by EntityManager
		private var _speedX : Number;
		private var _speedY : Number;
		private var _sprite : LiteSprite;
		public var active : Boolean = true;

		// if this is set, custom behaviors are run
		public var aiFunction : Function;
		// the AI routines might want access to the entity manager
		public var gfx:EntityManager;
		// AI needs to have access to time passing (in seconds)
		public var age:Number = 0;
		public var fireTime:Number = 0;
		public var fireDelayMin:Number = 1;
		public var fireDelayMax:Number = 6;
		// an array of points defining a movement path for the AI
		public var aiPathWaypoints:Array;
		// how fast we travel from one spline node to the next in seconds
		public var pathNodeTime:Number = 1;
		// these offsets are added to the sprite location
		// so that ships move around but eventually scroll offscreen
		public var aiPathOffsetX:Number = 0;
		public var aiPathOffsetY:Number = 0;
		// how much big the path is (max)
		public var aiPathSize:Number = 128;
		// how many different nodes in the path
		public var aiPathWaypointCount:int = 8;
		
		// collision detection
		public var isBullet:Boolean = false; // only these check collisions
		public var leavesTrail:Boolean = false; // creates particles as it moves
		public var collidemode:uint = 0; // 0=none, 1=sphere, 2=box, etc.
		public var collideradius:uint = 32; // used for sphere collision		
		// box collision is not implemented (yet)
		public var collidebox:Rectangle = new Rectangle(-16, -16, 32, 32);
		public var collidepoints:uint = 25; // score earned if destroyed
		public var touching:Entity; // what entity just hit us?

		public var owner:Entity; // so your own bullets don't hit you
		public var orbiting:Entity; // entities can orbit (circle) others
		public var orbitingDistance:Number; // how far in px from the orbit center
		
		// used for particle animation (in units per second)
		public var fadeAnim:Number = 0;
		public var zoomAnim:Number = 0;
		public var rotationSpeed:Number = 0;

		// used to mark whether or not this entity was 
		// freshly created or reused from an inactive one
		public var recycled:Boolean = false;
		
		public function Entity(gs:LiteSprite, myManager:EntityManager)
		{
			_sprite = gs;
			_speedX = 0.0;
			_speedY = 0.0;
			// we need a reference to the entity manager
			gfx = myManager;
			// we don't want everyone shooting the first frame
			fireTime = (gfx.fastRandom() * (fireDelayMax - fireDelayMin)) + fireDelayMin;
		}
		
		public function die() : void
		{
			// allow this entity to be reused by the entitymanager
			active = false;
			// skip all drawing and updating
			sprite.visible = false;
			// reset some things that might affect future reuses:
			leavesTrail = false;
			isBullet = false;
			touching = null;
			owner = null;
			age = 0;
			collidemode = 0;
		}
		
		public function get speedX() : Number 
		{
			return _speedX;
		}
		public function set speedX(sx:Number) : void 
		{
			_speedX = sx;
		}
		public function get speedY() : Number 
		{
			return _speedY;
		}
		public function set speedY(sy:Number) : void 
		{
			_speedY = sy;
		}
		public function get sprite():LiteSprite 
		{	
			return _sprite;
		}
		public function set sprite(gs:LiteSprite):void 
		{
			_sprite = gs;
		}

		// used for collision callback performed in GameActorpool
		public function colliding(checkme:Entity):Entity
		{
			if (collidemode == 1) // sphere
			{
				if (isCollidingSphere(checkme))
					return checkme;
			}
			return null;
		}
	
	// simple sphere to sphere collision
	public function isCollidingSphere(checkme:Entity):Boolean
	{
		// never collide with yourself
		if (this == checkme) return false;
		// only check if these shapes are collidable
		if (!collidemode || !checkme.collidemode) return false;
		// don't check your own bullets
		if (checkme.owner == this) return false;
		// don't check things on the same "team"
		if (checkme.owner == owner) return false;
		// don't check if no radius
		if (collideradius == 0 || checkme.collideradius == 0) return false;
		
		// this is the simpler way to do it, but it runs really slow
		// var dist:Number = Point.distance(sprite.position, checkme.sprite.position);
		// if (dist <= (collideradius+checkme.collideradius))
		
		// this looks wierd but is 6x faster than the above
		// see: http://www.emanueleferonato.com/2010/10/13/as3-geom-point-vs-trigonometry/
		if (((sprite.position.x - checkme.sprite.position.x) * 
			(sprite.position.x - checkme.sprite.position.x) +
			(sprite.position.y - checkme.sprite.position.y) *
			(sprite.position.y - checkme.sprite.position.y))
			<= 
			(collideradius+checkme.collideradius)*(collideradius+checkme.collideradius))
		{
			touching = checkme; // remember who hit us
			return true;
		}
		
		// default: too far away
		// trace("No collision. Dist = "+dist);
		return false;
		
	}


	// Calculates 2D cubic Catmull-Rom spline.
	// See http://www.mvps.org/directx/articles/catmull/ 
	public function spline (p0:Point, p1:Point, p2:Point, p3:Point, t:Number):Point 
	{
		return new Point (
			0.5 * ((2 * p1.x) +
			t * (( -p0.x + p2.x) +
			t * ((2 * p0.x -5 * p1.x +4 * p2.x -p3.x) +
			t * ( -p0.x +3 * p1.x -3 * p2.x +p3.x)))),
			0.5 * ((2 * p1.y) +
			t * (( -p0.y + p2.y) +
			t * ((2 * p0.y -5 * p1.y +4 * p2.y -p3.y) +
			t * (  -p0.y +3 * p1.y -3 * p2.y +p3.y))))
		);
	}
    
	// generate a random path
	public function generatePath():void
	{
		// v5
		// trace("Generating AI path");
		aiPathWaypoints = [];	
		var N:int = aiPathWaypointCount;
		for (var i:int = 0; i < N; i++) 
		{
			aiPathWaypoints.push (new Point (aiPathSize * Math.random (), aiPathSize * Math.random ()));
		}
	}
				
	// find the point on a spline at ratio (0 to 1)
	public function calculatePathPosition(ratio:Number = 0):Point
	{
		var i:int = int(ratio);
		var pointratio:Number = ratio - i;
		//trace(ratio + ' ratio = path point ' + i + ' segment ratio ' + pointratio);
		var p0:Point = aiPathWaypoints [(i -1 + aiPathWaypoints.length) % aiPathWaypoints.length];
		var p1:Point = aiPathWaypoints [i % aiPathWaypoints.length];
		var p2:Point = aiPathWaypoints [(i +1 + aiPathWaypoints.length) % aiPathWaypoints.length];
		var p3:Point = aiPathWaypoints [(i +2 + aiPathWaypoints.length) % aiPathWaypoints.length];
		// figure out current position
		var q:Point = spline (p0, p1, p2, p3, pointratio);
		return q;
	}
		
	///////////////////////
	// optional AI routines
	///////////////////////
	
	// we could optionally implement many different
	// versions of this routine with different randomness
	public function maybeShoot(bulletNum:int = 1, 
		delayMin:Number = NaN, 
		delayMax:Number = NaN):void
	{
		// is it time to shoot a bullet?
		if (fireTime < age)
		{
			// if delay parameters were not set, use class defaults
			if (isNaN(delayMin)) delayMin = fireDelayMin;
			if (isNaN(delayMax)) delayMax = fireDelayMax;
			
			// shoot one from the current location
			gfx.shootBullet(bulletNum, this);
			// randly choose the next time to shoot
			fireTime = age + (gfx.fastRandom() * (delayMax - delayMin)) + delayMin;
		}
	}
	
	// moves forward and points at current destination based on speed
	public function straightAI(seconds:Number):void
	{
		age += seconds;
		maybeShoot(1);
		sprite.rotation = gfx.pointAtRad(speedX, speedY)
			- (90*gfx.DEGREES_TO_RADIANS);
	}

	// a very simple up/down wobble movement
	public function wobbleAI(seconds:Number):void
	{
		age += seconds;
		maybeShoot(1);
		aiPathOffsetY = (Math.sin(age*2) / Math.PI) * 128;		
	}
	
	// simply point at the player: good for sentry guns
	public function sentryAI(seconds:Number):void
	{
		age += seconds;
		maybeShoot(3,3,6);
		if (gfx.thePlayer)
			sprite.rotation = gfx.pointAtRad(
				gfx.thePlayer.sprite.position.x - sprite.position.x,
				gfx.thePlayer.sprite.position.y - sprite.position.y) 
				- (90*gfx.DEGREES_TO_RADIANS);
	}
	
	// move around on a random spline path
	// in future versions, you could upgrade this function
	// to (instead of random) follow a predefined array of points
	// that were designed by hand in code (or even in the level editor!)
	public function droneAI(seconds:Number):void
	{
		//trace('droneAI');
		
		age += seconds;
		maybeShoot(1);

		// movement style inspired by Galaga, R-Type, Centipede
		// performed by easing through a catmull-rom spline curve
		// defined by an array of points
		if (aiPathWaypoints == null)
			generatePath();
			
		// how many spline nodes have we passed? (loops around to beginning)
		var pathProgress:Number = age / pathNodeTime;
			
		var newPos:Point = calculatePathPosition(pathProgress);

		// point in the correct direction
		//sprite.rotation = gfx.pointAtRad(newPos.x-sprite.position.x,newPos.y-sprite.position.y) - (90*gfx.DEGREES_TO_RADIANS);
		sprite.rotation = gfx.pointAtRad(newPos.x-aiPathOffsetX,newPos.y-aiPathOffsetY) - (90*gfx.DEGREES_TO_RADIANS);
		
		// change path offset location
		// this is added to the sprite scrolling location
		// so that ships eventually move offscreen
		// sprite.position.x = newPos.x;
		// sprite.position.y = newPos.y;
		
		aiPathOffsetX = newPos.x;
		aiPathOffsetY = newPos.x;
		
	}
	
	} // end class
} // end package